from contextlib import contextmanager

from gooey_gui import core
from gooey_gui.components import common as gui


class Modal:
    def __init__(self, title, key, padding=20, max_width=744):
        """
        :param title: title of the Modal shown in the h1
        :param key: unique key identifying this modal instance
        :param padding: padding of the content within the modal
        :param max_width: maximum width this modal should use
        """
        self.title = title
        self.padding = padding
        self.max_width = str(max_width) + "px"
        self.key = key

        self._container = None

    def is_open(self):
        return core.session_state.get(f"{self.key}-opened", False)

    def open(self):
        core.session_state[f"{self.key}-opened"] = True
        core.rerun()

    def close(self, rerun_condition=True):
        core.session_state[f"{self.key}-opened"] = False
        if rerun_condition:
            core.rerun()

    def empty(self):
        if self._container:
            self._container.empty()

    @contextmanager
    def container(self, **props):
        gui.html(
            f"""
        <style>
        .blur-background {{
            position: fixed;
            content: ' ';
            left: 0;
            right: 0;
            top: 0;
            bottom: 0;
            z-index: 9999;
            background-color: rgba(0, 0, 0, 0.5);
        }}
        .modal-parent {{
            position: fixed;
            top: 0;
            left: 0;
            width: 100vw;
            height: 100vh;
            z-index: 2000;
            display: flex;
            align-items: center;
            justify-content: center;
        }}
        .modal-container {{
            overflow-y: scroll;
            padding: 1.5rem;
            margin: auto;
            background: white;
            z-index: 3000;
            max-height: 80vh;
        }}
        </style>
        """
        )

        with gui.div(className="blur-background"):
            with gui.div(className="modal-parent"):
                container_class = "modal-container " + props.pop("className", "")
                self._container = gui.div(className=container_class, **props)

        with self._container:
            with gui.div(className="d-flex justify-content-between align-items-center"):
                if self.title:
                    gui.markdown(f"### {self.title}")
                else:
                    gui.div()

                close_ = gui.button(
                    "&#10006;",
                    type="tertiary",
                    key=f"{self.key}-close",
                    style={"padding": "0.375rem 0.75rem"},
                )
                if close_:
                    self.close()
            yield self._container



import hashlib
import threading
import typing
import uuid
from functools import wraps

import gooey_gui.components as gui
from .pubsub import realtime_pull, realtime_push
from .state import threadlocal, get_session_state

F = typing.TypeVar("F", bound=typing.Callable[..., typing.Any])


def cache_in_session_state(fn: F | None = None, key="__cache__") -> F:
    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            # hash the args and kwargs so they are not too long
            args_hash = hashlib.sha256(f"{args}{kwargs}".encode()).hexdigest()
            # create a readable cache key
            cache_key = fn.__name__ + ":" + args_hash
            session_state = get_session_state()
            try:
                # if the cache exists, return it
                result = session_state[key][cache_key]
            except KeyError:
                # otherwise, run the function and cache the result
                result = fn(*args, **kwargs)
                session_state.setdefault(key, {})[cache_key] = result
            return result

        return wrapper

    if fn:
        return decorator(fn)
    else:
        return decorator


def run_in_thread(
    fn: typing.Callable,
    *,
    args: typing.Sequence | None = None,
    kwargs: typing.Mapping | None = None,
    placeholder: str = "...",
    cache: bool = False,
    key: str | None = None,
    ex=60,
):
    if key is None:
        key = f"{run_in_thread.__name__}/{fn}"

    session_state = get_session_state()
    try:
        channel = session_state[key]
    except KeyError:
        channel = session_state[key] = (
            f"{run_in_thread.__name__}/{fn.__name__}/{uuid.uuid1()}"
        )

        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        def target():
            realtime_push(channel, dict(y=fn(*args, **kwargs)), ex=ex)

        threading.Thread(target=target).start()

    try:
        return session_state[channel]
    except KeyError:
        pass

    result = realtime_pull([channel])[0]
    if result and "y" in result:
        ret = result["y"]
        if cache:
            session_state[channel] = ret
        else:
            session_state.pop(key)
        return ret
    elif placeholder:
        gui.write(placeholder)


def use_state(initval, *, key: str | None = None, ex=60):
    if key is None:
        threadlocal.use_state_count += 1
        key = f"{use_state.__name__}/{threadlocal.use_state_count}"

    session_state = get_session_state()
    channel = session_state.setdefault(key, f"{use_state.__name__}/{uuid.uuid1()}")

    result = realtime_pull([channel])[0]
    if result and "y" in result:
        retval = session_state[channel] = result["y"]
    else:
        retval = session_state.setdefault(channel, initval)

    def set_state(val):
        realtime_push(channel, dict(y=val), ex=60)

    return retval, set_state



import hashlib
import json
import typing
from contextlib import contextmanager
from functools import lru_cache
from time import time

from loguru import logger

from .state import threadlocal

T = typing.TypeVar("T")


@lru_cache
def get_redis():
    import redis
    from decouple import config

    return redis.Redis.from_url(config("REDIS_URL", "redis://localhost:6379"))


def realtime_clear_subs():
    threadlocal.channels = []


def get_subscriptions() -> list[str]:
    try:
        return threadlocal.channels
    except AttributeError:
        threadlocal.channels = []
        return threadlocal.channels


def realtime_pull(channels: list[str]) -> list[typing.Any]:
    channels = [f"gooey-gui/state/{channel}" for channel in channels]
    threadlocal.channels = channels
    r = get_redis()
    out = [
        json.loads(value) if (value := r.get(channel)) else None for channel in channels
    ]
    return out


def realtime_push(channel: str, value: typing.Any = "ping", ex=None):
    from fastapi.encoders import jsonable_encoder

    channel = f"gooey-gui/state/{channel}"
    msg = json.dumps(jsonable_encoder(value))
    r = get_redis()
    r.set(channel, msg, ex=ex)
    r.publish(channel, json.dumps(time()))
    if isinstance(value, dict):
        run_status = value.get("__run_status")
        logger.info(f"publish {channel=} {run_status=}")
    else:
        logger.info(f"publish {channel=}")


@contextmanager
def realtime_subscribe(channel: str) -> typing.Generator:
    channel = f"gooey-gui/state/{channel}"
    r = get_redis()
    pubsub = r.pubsub()
    pubsub.subscribe(channel)
    logger.info(f"subscribe {channel=}")
    try:
        yield _realtime_sub_gen(channel, pubsub)
    finally:
        logger.info(f"unsubscribe {channel=}")
        pubsub.unsubscribe(channel)
        pubsub.close()


def _realtime_sub_gen(channel: str, pubsub: "redis.client.PubSub") -> typing.Generator:
    while True:
        message = pubsub.get_message(timeout=10)
        if not (message and message["type"] == "message"):
            continue
        r = get_redis()
        value = json.loads(r.get(channel))
        if isinstance(value, dict):
            run_status = value.get("__run_status")
            logger.info(f"realtime_subscribe: {channel=} {run_status=}")
        else:
            logger.info(f"realtime_subscribe: {channel=}")
        yield value


def md5_values(*values) -> str:
    strval = ".".join(map(repr, values))
    return hashlib.md5(strval.encode()).hexdigest()



from gooey_gui.components import common as gui


def url_button(url):
    gui.html(
        f"""
<a href='{url}' target='_blank'>
    <button type="button" class="btn btn-theme btn-tertiary">
        <i class="fa-regular fa-external-link-square"></i>
    </button>
</a>
        """
    )



import threading
import typing


threadlocal = threading.local()


def get_session_state() -> dict[str, typing.Any]:
    try:
        return threadlocal.session_state
    except AttributeError:
        threadlocal.session_state = {}
        return threadlocal.session_state


def set_session_state(state: dict[str, typing.Any]):
    threadlocal.session_state = state


def get_query_params() -> dict[str, str]:
    try:
        return threadlocal.query_params
    except AttributeError:
        threadlocal.query_params = {}
        return threadlocal.query_params


def set_query_params(params: dict[str, str]):
    threadlocal.query_params = params



import typing

from .exceptions import (
    RedirectException,
    QueryParamsRedirectException,
    StopException,
    RerunException,
    rerun,
    stop,
)
from .pubsub import (
    realtime_push,
    realtime_pull,
    realtime_subscribe,
    get_subscriptions,
    realtime_clear_subs,
    md5_values,
)
from .renderer import RenderTreeNode, NestingCtx, renderer, route
from .state import (
    get_session_state,
    set_session_state,
    get_query_params,
    set_query_params,
)
from .state_interactions import use_state, run_in_thread, cache_in_session_state

session_state: dict[str, typing.Any]


def __getattr__(name):
    if name == "session_state":
        return get_session_state()
    else:
        raise AttributeError(name)
